unit drwatch;

{ This is Version 1.1
 + Notification method by Troels S Eriksen. 
 + Will now attach to TMemo components etc as well
}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TFilesDroppedEvent = procedure(Sender: TObject; X,Y: Integer; Files: TStringList) of object;
  TDropWatch = class(TComponent)
  private
    { Private-Deklarationen }
    fEnabled: Boolean;
    fWatchControl: TWinControl;
    fOnFilesDropped: TFilesDroppedEvent;
    fHookRunning: Boolean;
    OldWndProc: TFarProc;
    NewWndProc: Pointer;
    procedure HookWndProc(var Message: TMessage);

  protected
    { Protected-Deklarationen }
    procedure SetEnabled(val: Boolean);
    procedure SetWatchControl(val: TWinControl);
    procedure Hook;
    procedure Unhook;
  public
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    { Public-Deklarationen }
    Constructor Create(anOwner: TComponent); override;
    Destructor  Destroy; override;
  published
    { Published-Deklarationen }
    Property WatchControl: TWinControl read fWatchControl write SetWatchControl stored True;
    Property Enabled: Boolean read fEnabled write SetEnabled stored True;
    Property OnFilesDropped: TFilesDroppedEvent read fOnFilesDropped write fOnFilesDropped stored True;
  end;

procedure Register;

implementation

Uses ShellApi;

Constructor TDropWatch.Create(anOwner: TComponent);
Begin
     inherited Create(anOwner);
     NewWndProc := MakeObjectInstance(HookWndProc);
     OldWndProc := NIL;
     fEnabled:=False;
     fHookRunning:=False;
     fWatchControl:=nil;
     fOnFilesDropped:=nil;
End;

Destructor TDropWatch.Destroy;
Begin
     Unhook;
     if assigned(NewWndProc) then
       FreeObjectInstance(NewWndProc);
     inherited Destroy;
End;

Procedure TDropWatch.SetEnabled(val: Boolean);
Begin
     If val=fEnabled Then Exit;
     fEnabled:=val;
     If fhookRunning Then DragAcceptFiles(fWatchControl.Handle, fEnabled);
End;

Procedure TDropWatch.SetWatchControl(val: TWincontrol);
Begin
     If val=fWatchControl Then Exit;
     Unhook;
     fWatchControl:=val;
     Hook;
End;

Procedure TDropWatch.Hook;
Begin
     { Hooks the WatchComponent. Makes sure,
       that there is a component assigned and
       we're NOT in design mode.
     }
     If (Assigned(FWatchControl)) And Not (csDesigning in ComponentState)
        { and IsWindow(fWatchControl.Handle) } Then Begin
          { This was borrowed from Brad Stowers' GrabBar component: }
          { Get the old window procedure via API call and store it. }
          OldWndProc := TFarProc(GetWindowLong(fWatchControl.Handle, GWL_WNDPROC));
          { Convert our object method into something Windows knows how to call }
          { Install it as the new Parent window procedure }
          SetWindowLong(fWatchControl.Handle, GWL_WNDPROC, LongInt(NewWndProc));
          DragAcceptFiles(fWatchControl.Handle, fEnabled);
          fHookRunning:=True;
     End Else fHookRunning:=False;
End;

procedure TDropWatch.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if (Operation=opRemove) and
     (AComponent=fWatchControl) then begin
    Enabled:=FALSE;
    WatchControl:=NIL;
  end;
end;


Procedure TDropWatch.Unhook;
Begin
     { Unhooks the WatchComponent. }
     If fhookRunning  {And Not (csDestroying in ComponentState)} Then
     Begin
          DragAcceptFiles(fWatchControl.Handle, False);
          if assigned(OldWndProc) then
            { Set back to original window procedure }
            SetWindowLong(fWatchControl.Handle, GWL_WNDPROC, LongInt(OldWndProc));
          { Reset variables to NIL }
          OldWndProc := NIL;
     End;
     fhookRunning:=False;
End;

procedure TDropWatch.HookWndProc(var Message: TMessage);
var
  DropPt     : TPoint;
  NumDropped : Integer;
  DroppedName: Array [0..255] of Char;
  I          : Integer;
  sl         : TStringList;
begin
  { If there's no fWatchControl, something has really gone wrong. }
  if fWatchControl = NIL then exit;
  with Message do begin
    { In GrabBar.pas, the original author, Brad Stowers writes: }
    { ALWAYS call the old window procedure so the parent can process its    }
    { messages.  Thanks to Gary Frerking for pointing me at CallWindowProc. }
    { I was trying to call the function directly, which died horribly.      }

    Result := CallWindowProc(OldWndProc, fWatchControl.Handle, Msg, wParam, lParam);

    if (Msg = WM_DESTROY) Then Unhook;

    { WatchControl gets a WM_DROPFILES message!}
    if (Msg = WM_DROPFILES) Then Begin
         { This method was borrowed from a Borland Pascal 7.0 Shell Demo }
         { Get the mouse cursors coordinates }

         DragQueryPoint(wParam, DropPt);

         { How many files were dropped? }

         NumDropped := DragQueryFile(wParam, $FFFFFFFF, nil, 0);

         { Get the filenames. }

         sl:=TStringList.Create;

         for I := 0 to NumDropped-1 do
         begin
           DragQueryFile(wParam, I, DroppedName, SizeOf(DroppedName));
           sl.Add(StrPas(DroppedName));
         end;

         If Assigned(fOnFilesDropped) Then { Anybody interested in the results? }
           fOnFilesDropped(Self,DropPt.X,DropPt.Y,sl);

         { We're finished with this message }
         DragFinish(wParam);
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('XWatch', [TDropWatch]);
end;

end.
